/* TODO: make create_output safe (buffer overflows) 
 */
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <grp.h>
#include <sys/types.h>
#include <pwd.h>

/* Maximum width of the terminal */
#define MAX_WIDTH 	134
#define VERSION_MAJOR	1
#define VERSION_MINOR	0
#define VERSION_RELEASE	"release"
#define DEFAULT_OUTPUT  "s"
#define DEFAULT_DELIM	" \t"

void print_line(char *msg);
void print_version(void);
void print_usage(void);
void print_array(int lines, char *msg[]);
void create_line(char *output, char *filename, int max_fname_size, 
		struct stat *st);
char *get_gid(int gid);
char *get_uid(int uid);
char *nice_size(long size);
char *nice_time(time_t t);
char *nice_datetime(time_t t);

int update_freq=1, whip_me_beat_me_make_me_y2k_compliant=0;
int dont_follow_symlinks=0;
int strip_path = 0;

char output_varibles[10];
char field_delim[6] = DEFAULT_DELIM;

char **files;
char **output;

int main(int argc, char *argv[])
{
	int i=0,s,count,maxfsize=0,l,rv;
	struct stat st;
	char *tmp;

	memset(output_varibles,0,sizeof(output_varibles));

	while((s = getopt(argc, argv, "sugaAmMcCnNVD:fF:YSh")) != -1) {
		switch(s) {
		case 'V':
			print_version();
			return 0;
		case 'h':
			print_usage();
			return 0;
		case 'f':
			strip_path = 1;
			break;
		case 'F':
			update_freq = atoi(optarg);
			if (update_freq < 1) {
				puts("Invalid update frequency");
				return 3;
			}
			break;
		case 'D':
			strncpy(field_delim,optarg,5);
			break;
		case 'S': dont_follow_symlinks = 1; break;
		case 'Y': whip_me_beat_me_make_me_y2k_compliant = 1; break;
		case '?': case ':': break;
		default: /* The display options */
			output_varibles[i++] = s;
		}
	}
	if (output_varibles[0] == 0) { /* Use defaults */
		strncpy(output_varibles,DEFAULT_OUTPUT,sizeof(output_varibles));
	} 

	count = argc - optind;
	if (count < 1) {
		print_usage();
		return 0;
	}

	files = (char **) malloc(sizeof(char *) * count);
	output = (char **) malloc(sizeof(char *) * count);

	/* Calculates the maximum filename size, among other things */
	for (i = optind; i < argc; i++) {
		files[i-optind] = strdup(argv[i]);
		if (strip_path && (tmp=strrchr(argv[i],'/')) != NULL)
			tmp++;
		else
			tmp=argv[i];
		l = strlen(tmp);

		if (l > maxfsize)
			maxfsize = l;

		output[i-optind] = (char *) malloc(sizeof(char) * (MAX_WIDTH+1));
		putc('\n',stdout); /* May aswell move the screen down now too */ 
	}

	while(1) {
		for (i = 0; i < count; i++) {
			if (dont_follow_symlinks)
				rv = lstat(files[i],&st);
			else
				rv = stat(files[i],&st);
		
			if (rv == -1) {
				strncpy(output[i],"Error stat(2)'ing file.",MAX_WIDTH);
				continue;
			}
			create_line(output[i],files[i],maxfsize,&st);
		}
		print_array(count,output);
		sleep(update_freq);
	}
	
	return 0;
}


void create_line(char *output, char *filename, int max_fname_size, struct stat *st)
{
	char tempstr[MAX_WIDTH+1],*f=NULL;
	int i;
	
	if (strip_path && (f = strrchr(filename,'/')) != NULL) 
		f++;
	else
		f = filename;

	sprintf(tempstr,"%-*s: ",max_fname_size,f);
		
	for(i = 0; output_varibles[i] != 0; i++) {
		
		switch(output_varibles[i]) {
		case 'u':
			strcat(tempstr,get_uid(st->st_uid));
			if (output_varibles[i+1] == 'g')
				strcat(tempstr,":");
			else
				strcat(tempstr,field_delim);
			continue; /* Just cause */
		case 's':
			if (S_ISDIR(st->st_mode))
				strcat(tempstr,"<dir>");
			else
				strcat(tempstr,nice_size(st->st_size));
			break;
		case 'g':	
			strcat(tempstr,get_gid(st->st_gid));
			break;
		case 'a':
			strcat(tempstr,nice_datetime(st->st_atime));
			break;
		case 'A':
			strcat(tempstr,nice_time(st->st_atime));
			break;
		case 'm':
			strcat(tempstr,nice_datetime(st->st_mtime));
			break;
		case 'M':
			strcat(tempstr,nice_time(st->st_mtime));
			break;
		case 'c':
			strcat(tempstr,nice_datetime(st->st_ctime));
			break;
		case 'C':	
			strcat(tempstr,nice_time(st->st_ctime));
			break;
		case 'n': 
			strcat(tempstr,nice_datetime(time(NULL)));
			break;
		case 'N': 
			strcat(tempstr,nice_time(time(NULL)));
			break;
		}
		strcat(tempstr,field_delim);
		
	}
	strcpy(output,tempstr);
}


char *nice_size(long size)
{
	static char buffer[21];
	char *n_nice;
	double n_bytes=size;

#define ONEMEG 1048576
#define ONEGIG (ONEMEG * 1024)
	if (n_bytes > ONEGIG) {
		n_bytes /= ONEGIG;
		n_nice = "GB";
	} else if (n_bytes > ONEMEG) {
		n_bytes /= ONEMEG;
		n_nice = "MB";
	} else if (n_bytes > 1024) {
		n_bytes /= 1024;
		n_nice = "KB";
	} else {
		snprintf(buffer,99,"%luB", size);
		return buffer;
	}
	snprintf(buffer,20,"%.2f%s",n_bytes,n_nice);
	return buffer;
}

char *nice_datetime(time_t t)
{
	struct tm *tinfo;
	static char buffer[21];

	tinfo = localtime(&t);

	/* Something to keep the paranoid happy */
	if (whip_me_beat_me_make_me_y2k_compliant) 
		snprintf(buffer,20,"%2d/%.2d/%.4d %.2d:%.2d:%.2d",
			tinfo->tm_mday,tinfo->tm_mon+1,(tinfo->tm_year + 1900),
			tinfo->tm_hour,tinfo->tm_min,tinfo->tm_sec);
	else 
		snprintf(buffer,20,"%2d/%.2d/%.2d %.2d:%.2d:%.2d",
			tinfo->tm_mday,tinfo->tm_mon+1,(tinfo->tm_year % 100),
			tinfo->tm_hour,tinfo->tm_min,tinfo->tm_sec);
	return buffer;
}
/* Call me for you know what ;) */
char *nice_time(time_t t)
{
	struct tm *tinfo;
	static char buffer[21];
	
	tinfo = localtime(&t);
	
	snprintf(buffer,20,"%.2d:%.2d:%.2d",
		tinfo->tm_hour,tinfo->tm_min,tinfo->tm_sec);
	return buffer;
}

char *get_uid(int uid) 
{
	struct passwd *pw;
	static char buffer[21];

	pw = getpwuid(uid);
	if (pw == NULL) 
		snprintf(buffer,20,"%d",uid);
	else
		snprintf(buffer,20,"%s",pw->pw_name);
	return buffer;
}

char *get_gid(int gid) 
{
	struct group *gp;
	static char buffer[21];

	gp = getgrgid(gid);
	if (gp == NULL) 
		snprintf(buffer,20,"%d",gid);
	else
		snprintf(buffer,20,"%s",gp->gr_name);
	return buffer;
}

/* Prints this array on the screen by first going up N lines, then
 * overwritting whats currently on those lines with the contents of
 * the array (wipes to End of line). N is the number of lines in the
 * array.
 */
void print_array(int lines, char *msg[])
{
	int i;

	printf("\033[%dA",lines); /* move up N lines */
	printf("\033[G"); /* move to beginning of line */
	for(i = 0; i < lines; i++) {
		printf("\033[J"); /* Clear to EOL */
		printf(msg[i]);
		printf("\n");
	}

}

/* Prints the line, on the current line, erasing whatever else is on it */
void print_line(char *msg)
{
	printf("\033[G"); /* Back to beginning of line */
	printf(msg);
	printf("\033[J"); /* Clear to end of line (?) */
}

void print_usage()
{
	puts("watchfile [-Vsugamch] [-F <num>] <files...>\n");
	puts("You may specify as many files as you wish, but if you specify");
	puts("more than the number of lines on your terminal, results are");
	puts("unspecified. Also, the more options you include, the wider");
	puts("the output. If the output is wider than your terminal, again,");
	puts("the results are unspecified. Usually you get scrolling if you");
	puts("exceed the size of your terminal.");
	puts("Options:");
	puts("   -V      Displays the program version");
	puts("   -F <num>  Update the display every <num> seconds");
	puts("             Defaults to 1 second");
	puts("   -f      Remove the path from the filename before displaying");
	puts("           it.");
	puts("   -h      Display this screen");
	puts("   -S      Don't follow symlinks, stat the actual link file");
	puts("   -Y      Display dates with 4 digit year instead of 2");
	puts("   -D <str>  Changes the default field deliminator from \" \\t\"");

	puts("             to <str>.");

	puts("The remaining options indicate what file stats are displayed");
	puts("NOTE: The order you specify dictates the order displayed in");
	puts("the output.");
	puts("   -s      Display file sizes");
	puts("   -u      Display file user owner");
	puts("   -g      Display file group owner");
	puts("   -a      Display last accessed date+time");
	puts("   -A      Same as above, but display only time");
	puts("   -m      Display last modified[1] date+time");
	puts("   -M      Same as above, but display only time");
	puts("   -c      Display last changed[2] date+time");
	puts("   -C      Same as above, but display only time");
	puts("   -n      Display CURRENT date+time");
	puts("   -N      Same as above, but display only time");
	puts("");
	puts("[1] Modified time is the time the file was last written to.");
	puts("[2] Changed time is the time the file info (eg uid, mode, etc");
	puts("    was last changed");
	puts("FOOTNOTE: If you want to see this entire help message a");
	puts("          screenfull at a time, use watchfile -h | more");
}

void print_version()
{
	printf("watchfile version %d.%d - %s\n",
		VERSION_MAJOR,VERSION_MINOR,VERSION_RELEASE);
	puts("Copyright 1999, Nick Clifford <zaf@nrc.co.nz>\n");
	puts("This file is licensed under the GNU General Public License.");
	puts("For more information on the licensing conditions and on the");
	puts("GPL, read the LICENSE file distributed with this application");
	puts("or visit this apps website.");
	puts("watchfile. Created by Nick 'Zaf' Clifford");
	puts("email: zaf@nrc.co.nz");
	puts("website: http://www.nrc.co.nz/Zaf/apps/watchfile");
}

